from __future__ import print_function
from requests import Session
import hashlib
import logging
import xml.dom.minidom

from .errors import (InvalidError, LoginError)

_LOGGER = logging.getLogger(__name__)


def get_text(nodelist):
    """Get the value from a text node."""
    value = []
    for node in nodelist:
        if node.nodeType == node.TEXT_NODE:
            value.append(node.data)
    return ''.join(value)


def get_node_value(node, name):
    return get_text(node.getElementsByTagName(name)[0].childNodes)


class Fritzhome(object):
    """Fritzhome object to communicate with the device."""

    _sid = None
    _session = None

    def __init__(self, host, user, password):
        self._host = host
        self._user = user
        self._password = password
        self._session = Session()

    def _request(self, url, params=None, timeout=10):
        """Send a request with paramerters."""
        rsp = self._session.get(url, params=params, timeout=timeout)
        rsp.raise_for_status()
        return rsp.text.strip()

    def _login_request(self, username=None, secret=None):
        """Send a login request with paramerters."""
        url = 'http://' + self._host + '/login_sid.lua'
        params = {}
        if username:
            params['username'] = username
        if secret:
            params['response'] = secret

        plain = self._request(url, params)
        dom = xml.dom.minidom.parseString(plain)
        sid = get_text(dom.getElementsByTagName('SID')[0].childNodes)
        challenge = get_text(
            dom.getElementsByTagName('Challenge')[0].childNodes)

        return (sid, challenge)

    def _logout_request(self):
        """Send a logout request."""
        _LOGGER.info('logout')
        url = 'http://' + self._host + '/login_sid.lua'
        params = {
            'security:command/logout': '1',
            'sid': self._sid
        }

        self._request(url, params)

    def _create_login_secret(self, challenge, password):
        """Create a login secret."""
        to_hash = (challenge + '-' + password).encode('UTF-16LE')
        hashed = hashlib.md5(to_hash).hexdigest()
        return '{0}-{1}'.format(challenge, hashed)

    def _aha_request(self, cmd, ain=None, param=None):
        """Send an AHA request."""
        url = 'http://' + self._host + '/webservices/homeautoswitch.lua'
        params = {
            'switchcmd': cmd,
            'sid': self._sid
        }
        if param:
            params['param'] = param
        if ain:
            params['ain'] = ain

        plain = self._request(url, params)
        if plain == 'inval':
            raise InvalidError
        return plain

    def login(self):
        """Login and get a valid session ID."""
        (sid, challenge) = self._login_request()
        if sid == '0000000000000000':
            secret = self._create_login_secret(challenge, self._password)
            (sid2, challenge) = self._login_request(username=self._user,
                                                    secret=secret)
            if sid2 == '0000000000000000':
                _LOGGER.warning("login failed %s", sid2)
                raise LoginError(self._user)
        self._sid = sid2

    def logout(self):
        """Logout."""
        self._logout_request()
        self._sid = None

    def get_device_elements(self):
        plain = self._aha_request('getdevicelistinfos')
        dom = xml.dom.minidom.parseString(plain)
        return dom.getElementsByTagName("device")

    def get_device_element(self, ain):
        elements = self.get_device_elements()
        for element in elements:
            if element.getAttribute('identifier') == ain:
                return element
        return None

    def get_devices(self):
        """Get the list of all known devices."""
        devices = []
        for element in self.get_device_elements():
            device = FritzhomeDevice(self, node=element)
            devices.append(device)
        return devices

    def get_device_by_ain(self, ain):
        """Returns a device specified by the AIN."""
        devices = self.get_devices()
        for device in devices:
            if device.ain == ain:
                return device

    def get_device_present(self, ain):
        plain = self._aha_request('getswitchpresent', ain=ain)
        return bool(int(plain))

    def get_device_name(self, ain):
        plain = self._aha_request('getswitchname', ain=ain)
        return plain

    def get_switch_state(self, ain):
        plain = self._aha_request('getswitchstate', ain=ain)
        return bool(int(plain))

    def set_switch_state_on(self, ain):
        plain = self._aha_request('setswitchon', ain=ain)
        return bool(int(plain))

    def set_switch_state_off(self, ain):
        plain = self._aha_request('setswitchoff', ain=ain)
        return bool(int(plain))

    def set_switch_state_toggle(self, ain):
        plain = self._aha_request('setswitchtoggle', ain=ain)
        return bool(int(plain))

    def get_switch_power(self, ain):
        plain = self._aha_request('getswitchpower', ain=ain)
        return int(plain)

    def get_switch_energy(self, ain):
        plain = self._aha_request('getswitchenergy', ain=ain)
        return int(plain)

    def get_temperature(self, ain):
        plain = self._aha_request('gettemperature', ain=ain)
        return float(int(plain) / 10.0)

    def get_target_temperature(self, ain):
        plain = self._aha_request('gethkrtsoll', ain=ain)
        return ((float(plain) - 16) / 2 + 8)

    def set_target_temperature(self, ain, temperature):
        param = 16 + ((temperature - 8) * 2)

        r = range(16, 56)
        if param < r[0]:
            param = 253
        elif param > r[-1]:
            param = 254
        self._aha_request('sethkrtsoll', ain=ain, param=int(param))

    def get_comfort_temperature(self, ain):
        plain = self._aha_request('gethkrkomfort', ain=ain)
        return ((float(plain) - 16) / 2 + 8)

    def get_eco_temperature(self, ain):
        plain = self._aha_request('gethkrabsenk', ain=ain)
        return ((float(plain) - 16) / 2 + 8)


class FritzhomeDevice(object):

    ALARM_MASK = 0x010
    UNKNOWN_MASK = 0x020
    THERMOSTAT_MASK = 0x040
    POWER_METER_MASK = 0x080
    TEMPERATURE_MASK = 0x100
    SWITCH_MASK = 0x200
    DECT_REPEATER_MASK = 0x400
    UNKNOWN2_MASK = 0x800

    ain = None
    _id = None
    manufacturer = None
    productname = None
    actual_temperature = None
    target_temperature = None
    eco_temperature = None
    comfort_temperature = None
    lock = None
    device_lock = None
    error_code = None
    battery_low = None
    switch_state = None
    switch_mode = None
    power = None
    energy = None
    offset = None
    temperature = None

    def __init__(self, fritz=None, node=None, *args, **kwargs):
        if fritz is not None:
            self._fritz = fritz
        if node is not None:
            self._update_from_node(node)

    def _update_from_node(self, node):
        _LOGGER.debug(node.toprettyxml())
        self.ain = node.getAttribute("identifier")
        self.id = node.getAttribute("id")
        self._functionsbitmask = int(node.getAttribute("functionbitmask"))
        self.fw_version = node.getAttribute("fwversion")
        self.manufacturer = node.getAttribute("manufacturer")
        self.productname = node.getAttribute("productname")

        self.name = get_node_value(node, 'name')
        self.present = bool(int(get_node_value(node, 'present')))

        if self.present is False:
            return

        if self.has_thermostat:
            n = node.getElementsByTagName('hkr')[0]
            self.actual_temperature = int(get_node_value(n, 'tist')) / 2
            self.target_temperature = int(get_node_value(n, 'tsoll')) / 2
            self.eco_temperature = int(get_node_value(n, 'absenk')) / 2
            self.comfort_temperature = int(get_node_value(n, 'komfort')) / 2

            # optional value
            try:
                self.device_lock = bool(int(get_node_value(n, 'devicelock')))
            except IndexError:
                pass

            try:
                self.lock = bool(int(get_node_value(n, 'lock')))
            except IndexError:
                pass

            try:
                self.error_code = int(get_node_value(n, 'errorcode'))
            except IndexError:
                pass

            try:
                self.battery_low = bool(int(get_node_value(n, 'batterylow')))
            except IndexError:
                pass

        if self.has_switch:
            n = node.getElementsByTagName('switch')[0]
            self.switch_state = bool(int(get_node_value(n, 'state')))
            self.switch_mode = get_node_value(n, 'mode')
            self.lock = bool(get_node_value(n, 'lock'))
            # optional value
            try:
                self.device_lock = bool(int(get_node_value(n, 'devicelock')))
            except IndexError:
                pass

        if self.has_powermeter:
            n = node.getElementsByTagName('powermeter')[0]
            self.power = int(get_node_value(n, 'power'))
            self.energy = int(get_node_value(n, 'energy'))

        if self.has_temperature_sensor:
            n = node.getElementsByTagName('temperature')[0]
            self.offset = int(get_node_value(n, 'offset')) / 10
            self.temperature = int(get_node_value(n, 'celsius')) / 10

    def __repr__(self):
        return '{} {} {} {}'.format(self.ain, self._id,
                                    self.manufacturer, self.productname)

    def update(self):
        node = self._fritz.get_device_element(self.ain)
        self._update_from_node(node)

    @property
    def has_alarm(self):
        return bool(self._functionsbitmask & self.ALARM_MASK)

    @property
    def has_thermostat(self):
        return bool(self._functionsbitmask & self.THERMOSTAT_MASK)

    @property
    def has_powermeter(self):
        return bool(self._functionsbitmask & self.POWER_METER_MASK)

    @property
    def has_temperature_sensor(self):
        return bool(self._functionsbitmask & self.TEMPERATURE_MASK)

    @property
    def has_switch(self):
        return bool(self._functionsbitmask & self.SWITCH_MASK)

    @property
    def has_repeater(self):
        return bool(self._functionsbitmask & self.DECT_REPEATER_MASK)

    def get_present(self):
        return self._fritz.get_device_present(self.ain)

    def get_switch_state(self):
        return self._fritz.get_switch_state(self.ain)

    def set_switch_state_on(self):
        return self._fritz.set_switch_state_on(self.ain)

    def set_switch_state_off(self):
        return self._fritz.set_switch_state_off(self.ain)

    def set_switch_state_toggle(self):
        return self._fritz.set_switch_state_toggle(self.ain)

    def get_switch_power(self):
        return self._fritz.get_switch_power(self.ain)

    def get_switch_energy(self):
        return self._fritz.get_switch_energy(self.ain)

    def get_temperature(self):
        return self._fritz.get_temperature(self.ain)

    def get_target_temperature(self):
        return self._fritz.get_target_temperature(self.ain)

    def set_target_temperature(self, temperature):
        return self._fritz.set_target_temperature(self.ain, temperature)

    def get_comfort_temperature(self):
        return self._fritz.get_comfort_temperature(self.ain)

    def get_eco_temperature(self):
        return self._fritz.get_eco_temperature(self.ain)



#!/usr/bin/env python

from __future__ import print_function
import logging
import argparse

try:
    from version import __version__
except ImportError:
    __version__ = 'dev'

from pyfritzhome import Fritzhome

_LOGGER = logging.getLogger(__name__)


def list_all(fritz, args):
    """Command that prints all device information."""
    devices = fritz.get_devices()

    for device in devices:
        print('#' * 30)
        print('name=%s' % device.name)
        print('  ain=%s' % device.ain)
        print('  id=%s' % device.id)
        print('  productname=%s' % device.productname)
        print('  manufacturer=%s' % device.manufacturer)
        print("  present=%s" % device.present)

        if device.present is False:
            continue

        if device.has_switch:
            print(" Switch:")
            print("  switch_state=%s" % device.switch_state)
        if device.has_switch:
            print(" Powermeter:")
            print("  switch_power=%s" % device.power)
            print("  switch_energy=%s" % device.energy)
        if device.has_temperature_sensor:
            print(" Temperature:")
            print("  temperature=%f" % device.temperature)
            print("  offset=%f" % device.offset)
        if device.has_thermostat:
            print(" Thermostat:")
            print("  battery_low=%s" % device.battery_low)
            print("  actual=%s" % device.actual_temperature)
            print("  target=%s" % device.target_temperature)
            print("  comfort=%s" % device.comfort_temperature)
            print("  eco=%s" % device.eco_temperature)


def device_name(fritz, args):
    """Command that prints the device name."""
    print(fritz.get_actor_name(args.ain))


def device_presence(fritz, args):
    """Command that prints the device presence."""
    print(int(fritz.get_actor_present(args.ain)))


def switch_get(fritz, args):
    """Command that get the device switch state."""
    print(fritz.get_switch_state(args.ain))


def switch_on(fritz, args):
    """Command that set the device switch state to on."""
    fritz.set_switch_state_on(args.ain)


def switch_off(fritz, args):
    """Command that set the device switch state to off."""
    fritz.set_switch_state_off(args.ain)


def switch_toggle(fritz, args):
    """Command that toggles the device switch state."""
    fritz.set_switch_state_toggle(args.ain)


def main(args=None):
    """The main function."""
    parser = argparse.ArgumentParser(
        description='Fritz!Box Smarthome CLI tool.')
    parser.add_argument('-v', action='store_true', dest='verbose',
                        help='be more verbose')
    parser.add_argument('-f', '--fritzbox', type=str, dest='host',
                        help='Fritz!Box IP address', default='fritz.box')
    parser.add_argument('-u', '--user', type=str, dest='user',
                        help='Username')
    parser.add_argument('-p', '--password', type=str, dest='password',
                        help='Username')
    parser.add_argument('-a', '--ain', type=str, dest='ain',
                        help='Actor Identification', default=None)
    parser.add_argument('-V', '--version', action='version',
                        version='{version}'.format(version=__version__),
                        help='Print version')

    _sub = parser.add_subparsers(title='Commands')

    # list all devices
    subparser = _sub.add_parser('list', help='List all available devices')
    subparser.set_defaults(func=list_all)

    # device
    subparser = _sub.add_parser('device', help='Device/Actor commands')
    _sub_switch = subparser.add_subparsers()

    # device name
    subparser = _sub_switch.add_parser('name', help='get the device name')
    subparser.add_argument('ain', type=str, metavar="AIN",
                           help='Actor Identification')
    subparser.set_defaults(func=device_name)

    # device presence
    subparser = _sub_switch.add_parser('present',
                                       help='get the device presence')
    subparser.add_argument('ain', type=str, metavar="AIN",
                           help='Actor Identification')
    subparser.set_defaults(func=device_presence)

    # switch
    subparser = _sub.add_parser('switch', help='Switch commands')
    _sub_switch = subparser.add_subparsers()

    # switch get
    subparser = _sub_switch.add_parser('get', help='get state')
    subparser.add_argument('ain', type=str, metavar="AIN",
                           help='Actor Identification')
    subparser.set_defaults(func=switch_get)

    # switch on
    subparser = _sub_switch.add_parser('on', help='set on state')
    subparser.add_argument('ain', type=str, metavar="AIN",
                           help='Actor Identification')
    subparser.set_defaults(func=switch_on)

    # switch off
    subparser = _sub_switch.add_parser('off', help='set off state')
    subparser.add_argument('ain', type=str, metavar="AIN",
                           help='Actor Identification')
    subparser.set_defaults(func=switch_off)

    # switch toggle
    subparser = _sub_switch.add_parser('toggle', help='set off state')
    subparser.add_argument('ain', type=str, metavar="AIN",
                           help='Actor Identification')
    subparser.set_defaults(func=switch_toggle)

    args = parser.parse_args(args)

    logging.basicConfig()
    if args.verbose:
        logging.getLogger('pyfritzhome').setLevel(logging.DEBUG)

    fritzbox = None
    try:
        fritzbox = Fritzhome(host=args.host, user=args.user,
                             password=args.password)
        fritzbox.login()
        args.func(fritzbox, args)
    finally:
        if fritzbox is not None:
            fritzbox.logout()


if __name__ == '__main__':
    main()



#!/usr/bin/env python

from nose.tools import eq_, raises
from mock import MagicMock

from requests import Response
from pyfritzhome import (Fritzhome, InvalidError, LoginError)

from .elements import (device_list_xml, login_rsp_without_valid_sid,
                       login_rsp_with_valid_sid)


class TestFritzhome(object):

    @raises(LoginError)
    def test_login_fail(self):
        mock = MagicMock()
        mock.side_effect = [
            login_rsp_without_valid_sid,
            login_rsp_without_valid_sid,
        ]

        fritz = Fritzhome('10.0.0.1', 'user', 'pass')
        fritz._request = mock
        fritz.login()

    def test_login(self):
        mock = MagicMock()
        mock.side_effect = [
            login_rsp_without_valid_sid,
            login_rsp_with_valid_sid,
        ]

        fritz = Fritzhome('10.0.0.1', 'user', 'pass')
        fritz._request = mock
        fritz.login()

    def test_logout(self):
        mock = MagicMock()

        fritz = Fritzhome('10.0.0.1', 'user', 'pass')
        fritz._request = mock

        fritz.logout()
        fritz._request.assert_called_with(
            'http://10.0.0.1/login_sid.lua',
            {'sid': None, 'security:command/logout': '1'})

    def test_aha_request(self):
        mock = MagicMock()

        fritz = Fritzhome('10.0.0.1', 'user', 'pass')
        fritz._request = mock

        fritz._aha_request(cmd='testcmd')
        fritz._request.assert_called_with(
            'http://10.0.0.1/webservices/homeautoswitch.lua',
            {'sid': None, 'switchcmd': 'testcmd'})

        fritz._aha_request(cmd='testcmd', ain='1')
        fritz._request.assert_called_with(
            'http://10.0.0.1/webservices/homeautoswitch.lua',
            {'sid': None, 'switchcmd': 'testcmd', 'ain': '1'})

        fritz._aha_request(cmd='testcmd', ain='1', param={'a': '1', 'b': '2'})
        fritz._request.assert_called_with(
            'http://10.0.0.1/webservices/homeautoswitch.lua',
            {'sid': None, 'switchcmd': 'testcmd', 'ain': '1',
             'param': {'a': '1', 'b': '2'}})

    @raises(InvalidError)
    def test_aha_request_invalid(self):
        mock = MagicMock()
        mock.side_effect = [
            'inval',
        ]

        fritz = Fritzhome('10.0.0.1', 'user', 'pass')
        fritz._request = mock

        fritz._aha_request(cmd='estcmd')

    def test_get_device_element(self):
        mock = MagicMock()
        mock.side_effect = [
            device_list_xml,
            device_list_xml,
            device_list_xml,
        ]

        fritz = Fritzhome('10.0.0.1', 'user', 'pass')
        fritz._request = mock
        element = fritz.get_device_element('08761 0000434')
        eq_(element.getAttribute('identifier'), '08761 0000434')
        eq_(element.getAttribute('fwversion'), '03.33')

        element = fritz.get_device_element('08761 1048079')
        eq_(element.getAttribute('identifier'), '08761 1048079')
        eq_(element.getAttribute('fwversion'), '03.44')

        element = fritz.get_device_element('unknown')
        eq_(element, None)

    def test_get_device_by_ain(self):
        mock = MagicMock()
        mock.side_effect = [
            device_list_xml,
            device_list_xml,
        ]

        fritz = Fritzhome('10.0.0.1', 'user', 'pass')
        fritz._request = mock
        device = fritz.get_device_by_ain('08761 0000434')
        eq_(device.ain, '08761 0000434')

    def test_aha_get_devices(self):

        mock = MagicMock()
        mock.side_effect = [
            device_list_xml,
        ]

        fritz = Fritzhome('10.0.0.1', 'user', 'pass')
        fritz._request = mock
        devices = fritz.get_devices()
        eq_(devices[0].name, 'Steckdose')
        eq_(devices[1].name, 'FRITZ!DECT Rep 100 #1')

    def test_get_device_name(self):

        mock = MagicMock()
        mock.side_effect = [
            'testname'
        ]

        fritz = Fritzhome('10.0.0.1', 'user', 'pass')
        fritz._request = mock

        eq_(fritz.get_device_name(ain='1234'), 'testname')

        fritz._request.assert_called_with(
            'http://10.0.0.1/webservices/homeautoswitch.lua',
            {'sid': None, 'ain': '1234', 'switchcmd': 'getswitchname'})

    def test_set_target_temperature(self):
        mock = MagicMock()

        fritz = Fritzhome('10.0.0.1', 'user', 'pass')
        fritz._request = mock

        fritz.set_target_temperature('1', 25.5)
        fritz._request.assert_called_with(
            'http://10.0.0.1/webservices/homeautoswitch.lua',
            {'sid': None, 'ain': '1', 'switchcmd': 'sethkrtsoll', 'param': 51})

        fritz.set_target_temperature('1', 0.0)
        fritz._request.assert_called_with(
            'http://10.0.0.1/webservices/homeautoswitch.lua',
            {'sid': None, 'ain': '1', 'switchcmd': 'sethkrtsoll', 'param': 253})

        fritz.set_target_temperature('1', 32.0)
        fritz._request.assert_called_with(
            'http://10.0.0.1/webservices/homeautoswitch.lua',
            {'sid': None, 'ain': '1', 'switchcmd': 'sethkrtsoll', 'param': 254})
